"""
Convenience interface for NSDictionary/NSMutableDictionary
"""
__all__ = ()

import collections.abc

from objc._convenience import addConvenienceForClass, container_wrap
from objc._convenience_mapping import addConvenienceForBasicMapping
from objc._objc import lookUpClass

NSDictionary = lookUpClass("NSDictionary")
NSMutableDictionary = lookUpClass("NSMutableDictionary")

addConvenienceForBasicMapping("NSDictionary", True)
addConvenienceForBasicMapping("NSMutableDictionary", False)


def _all_contained_in(inner, outer):
    """
    Return True iff all items in ``inner`` are also in ``outer``.
    """
    for v in inner:
        if v not in outer:
            return False

    return True


def nsdict__len__(self):
    return self.count()


def nsdict__iter__(self):
    return iter(self.keyEnumerator())


class nsdict_view(collections.abc.Set):
    __slots__ = ()

    def __eq__(self, other):
        if not isinstance(other, collections.abc.Set):
            return NotImplemented

        if len(self) == len(other):
            return _all_contained_in(self, other)

        else:
            return False

    def __ne__(self, other):
        if not isinstance(other, collections.abc.Set):
            return NotImplemented

        if len(self) == len(other):
            return not _all_contained_in(self, other)

        else:
            return True

    def __lt__(self, other):
        if not isinstance(other, collections.abc.Set):
            return NotImplemented

        if len(self) < len(other):
            return _all_contained_in(self, other)

        else:
            return False

    def __le__(self, other):
        if not isinstance(other, collections.abc.Set):
            return NotImplemented

        if len(self) <= len(other):
            return _all_contained_in(self, other)

        else:
            return False

    def __gt__(self, other):
        if not isinstance(other, collections.abc.Set):
            return NotImplemented

        if len(self) > len(other):
            return _all_contained_in(other, self)

        else:
            return False

    def __ge__(self, other):
        if not isinstance(other, collections.abc.Set):
            return NotImplemented

        if len(self) >= len(other):
            return _all_contained_in(other, self)

        else:
            return False

    def __and__(self, other):
        if not isinstance(other, collections.abc.Set):
            return NotImplemented
        result = set(self)
        result.intersection_update(other)
        return result

    def __rand__(self, other):
        if not isinstance(other, collections.abc.Set):
            return NotImplemented
        result = set(self)
        result.intersection_update(other)
        return result

    def __or__(self, other):
        if not isinstance(other, collections.abc.Set):
            return NotImplemented
        result = set(self)
        result.update(other)
        return result

    def __ror__(self, other):
        if not isinstance(other, collections.abc.Set):
            return NotImplemented
        result = set(self)
        result.update(other)
        return result

    def __sub__(self, other):
        if not isinstance(other, collections.abc.Set):
            return NotImplemented
        result = set(self)
        result.difference_update(other)
        return result

    def __rsub__(self, other):
        if not isinstance(other, collections.abc.Set):
            return NotImplemented
        result = set(other)
        result.difference_update(self)
        return result

    def __xor__(self, other):
        if not isinstance(other, collections.abc.Set):
            return NotImplemented
        result = set(self)
        result.symmetric_difference_update(other)
        return result

    def __rxor__(self, other):
        if not isinstance(other, collections.abc.Set):
            return NotImplemented
        result = set(self)
        result.symmetric_difference_update(other)
        return result


class nsdict_keys(nsdict_view):
    __slots__ = ("__value",)

    def __init__(self, value):
        self.__value = value

    def __repr__(self):
        keys = list(self.__value)

        return f"<nsdict_keys({keys})>"

    def __len__(self):
        return len(self.__value)

    def __iter__(self):
        return iter(self.__value)

    def __contains__(self, value):
        return value in self.__value


class nsdict_values(nsdict_view):
    __slots__ = ("__value",)

    def __init__(self, value):
        self.__value = value

    def __repr__(self):
        values = list(self)
        values.sort()

        return f"<nsdict_values({values})>"

    def __len__(self):
        return len(self.__value)

    def __iter__(self):
        return iter(self.__value.objectEnumerator())

    def __contains__(self, value):
        for v in iter(self):
            if value == v:
                return True
        return False


class nsdict_items(nsdict_view):
    __slots__ = ("__value",)

    def __init__(self, value):
        self.__value = value

    def __repr__(self):
        values = list(self)
        values.sort()

        return f"<nsdict_items({values})>"

    def __len__(self):
        return len(self.__value)

    def __iter__(self):
        for k in self.__value:
            yield (k, self.__value[k])

    def __contains__(self, value):
        for v in iter(self):
            if value == v:
                return True
        return False


collections.abc.KeysView.register(nsdict_keys)
collections.abc.ValuesView.register(nsdict_values)
collections.abc.ItemsView.register(nsdict_items)

collections.abc.Mapping.register(NSDictionary)
collections.abc.MutableMapping.register(NSMutableDictionary)


def nsdict_fromkeys(cls, keys, value=None):
    keys = [container_wrap(k) for k in keys]
    values = [container_wrap(value)] * len(keys)

    return cls.dictionaryWithObjects_forKeys_(values, keys)


def nsmutabledict_fromkeys(cls, keys, value=None):
    value = container_wrap(value)

    result = cls.alloc().init()
    for k in keys:
        result[container_wrap(k)] = value
    return result


def nsdict_new(cls, *args, **kwds):
    if len(args) == 0:
        pass

    elif len(args) == 1:
        d = {}
        if isinstance(args[0], collections.abc.Mapping):
            items = args[0].items()
        else:
            items = args[0]
        for k, v in items:
            d[container_wrap(k)] = container_wrap(v)

        for k, v in kwds.items():
            d[container_wrap(k)] = container_wrap(v)

        return cls.dictionaryWithDictionary_(d)

    else:
        raise TypeError(f"dict expected at most 1 arguments, got {len(args)}")

    if kwds:
        d = {}
        for k, v in kwds.items():
            d[container_wrap(k)] = container_wrap(v)

        return cls.dictionaryWithDictionary_(d)

    return cls.dictionary()


def nsdict__eq__(self, other):
    if not isinstance(other, collections.abc.Mapping):
        return False

    return self.isEqualToDictionary_(other)


def nsdict__ne__(self, other):
    return not nsdict__eq__(self, other)


def nsdict__lt__(self, other):
    return NotImplemented


def nsdict__le__(self, other):
    return NotImplemented


def nsdict__ge__(self, other):
    return NotImplemented


def nsdict__gt__(self, other):
    return NotImplemented


addConvenienceForClass(
    "NSDictionary",
    (
        ("keys", lambda self: nsdict_keys(self)),
        ("values", lambda self: nsdict_values(self)),
        ("items", lambda self: nsdict_items(self)),
    ),
)


addConvenienceForClass(
    "NSDictionary",
    (
        ("__new__", staticmethod(nsdict_new)),
        ("fromkeys", classmethod(nsdict_fromkeys)),
        ("__eq__", nsdict__eq__),
        ("__ne__", nsdict__ne__),
        ("__lt__", nsdict__lt__),
        ("__le__", nsdict__le__),
        ("__gt__", nsdict__gt__),
        ("__ge__", nsdict__ge__),
        ("__len__", nsdict__len__),
        ("__iter__", nsdict__iter__),
    ),
)


addConvenienceForClass(
    "NSMutableDictionary",
    (
        ("__new__", staticmethod(nsdict_new)),
        ("fromkeys", classmethod(nsmutabledict_fromkeys)),
        ("clear", lambda self: self.removeAllObjects()),
    ),
)



from objc import _objc

__all__ = []


class OC_PythonFloat(float):
    __slots__ = ("__pyobjc_object__",)

    def __new__(cls, obj, value):
        self = float.__new__(cls, value)
        self.__pyobjc_object__ = obj
        return self

    __class__ = property(lambda self: self.__pyobjc_object__.__class__)

    def __getattr__(self, attr):
        return getattr(self.__pyobjc_object__, attr)

    def __reduce__(self):
        return (float, (float(self),))


class OC_PythonLong(int):
    def __new__(cls, obj, value):
        self = int.__new__(cls, value)
        self.__pyobjc_object__ = obj
        return self

    __class__ = property(lambda self: self.__pyobjc_object__.__class__)

    def __getattr__(self, attr):
        return getattr(self.__pyobjc_object__, attr)

    # The long type doesn't support __slots__ on subclasses, fake
    # one part of the effect of __slots__: don't allow setting of attributes.
    def __setattr__(self, attr, value):
        if attr != "__pyobjc_object__":
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute '{attr}')"
            )
        self.__dict__["__pyobjc_object__"] = value

    def __reduce__(self):
        return (int, (int(self),))


NSNumber = _objc.lookUpClass("NSNumber")
NSDecimalNumber = _objc.lookUpClass("NSDecimalNumber")


def numberWrapper(obj):
    if isinstance(obj, NSDecimalNumber):
        return obj

    try:
        tp = obj.objCType()
    except AttributeError:
        import warnings

        warnings.warn(
            f"NSNumber instance doesn't implement objCType? {obj!r}", RuntimeWarning
        )
        return obj

    if tp in b"qQLfd":
        if tp == b"q":
            return OC_PythonLong(obj, obj.longLongValue())
        elif tp in b"QL":
            return OC_PythonLong(obj, obj.unsignedLongLongValue())
        else:
            return OC_PythonFloat(obj, obj.doubleValue())
    else:
        return OC_PythonLong(obj, obj.longValue())


_objc.options._nsnumber_wrapper = numberWrapper



"""
Specific support for NSData.

NSData needs to be handles specially for correctness reasons,
and is therefore in the core instead of the Foundation
framework wrappers.
"""
from objc._convenience import addConvenienceForClass
from objc._objc import registerMetaDataForSelector
import sys
import operator

registerMetaDataForSelector(
    b"NSData",
    b"dataWithBytes:length:",
    {
        "arguments": {
            2: {"type": b"^v", "type_modifier": b"n", "c_array_length_in_arg": 3}
        }
    },
)
registerMetaDataForSelector(
    b"NSData",
    b"initWithBytes:length:",
    {
        "arguments": {
            2: {"type": b"^v", "type_modifier": b"n", "c_array_length_in_arg": 3}
        }
    },
)
registerMetaDataForSelector(
    b"NSData",
    b"replaceBytesInRange:withBytes:length:",
    {
        "arguments": {
            3: {"type": b"^v", "type_modifier": b"n", "c_array_length_in_arg": 4}
        }
    },
)


def nsdata__new__(cls, value=None):
    if value is None:
        return cls.data()

    else:
        try:
            view = memoryview(value)
        except TypeError:
            view = bytes(value)

        return cls.dataWithBytes_length_(view, len(view))


def nsdata__str__(self):
    if len(self) == 0:
        return str(b"")
    return str(self.bytes().tobytes())


def nsdata__bytes__(self):
    return bytes(self.bytes())


def nsdata__getitem__(self, item):
    buff = self.bytes()
    return buff[item]


# XXX: The methods below are a bit too simplistic,
#      use NSData methods were applicable
def nsdata_capitalize(self):
    return bytes(self.bytes()).capitalize()


def nsdata_center(self, *args, **kwds):
    return bytes(self.bytes()).center(*args, **kwds)


def nsdata_count(self, *args, **kwds):
    return bytes(self.bytes()).count(*args, **kwds)


def nsdata_decode(self, *args, **kwds):
    return bytes(self.bytes()).decode(*args, **kwds)


def nsdata_endswith(self, *args, **kwds):
    return bytes(self.bytes()).endswith(*args, **kwds)


def nsdata_expandtabs(self, *args, **kwds):
    return bytes(self.bytes()).expandtabs(*args, **kwds)


def nsdata_find(self, *args, **kwds):
    return bytes(self.bytes()).find(*args, **kwds)


def nsdata_fromhex(cls, *args, **kwds):
    return cls.dataWithData_(bytes.fromhex(*args, **kwds))


def nsdata_hex(self, *args, **kwds):
    return memoryview(self.bytes()).hex(*args, **kwds)


def nsdata_index(self, *args, **kwds):
    return bytes(self.bytes()).index(*args, **kwds)


def nsdata_isalnum(self, *args, **kwds):
    return bytes(self.bytes()).isalnum(*args, **kwds)


def nsdata_isalpha(self, *args, **kwds):
    return bytes(self.bytes()).isalpha(*args, **kwds)


def nsdata_isdigit(self, *args, **kwds):
    return bytes(self.bytes()).isdigit(*args, **kwds)


def nsdata_islower(self, *args, **kwds):
    return bytes(self.bytes()).islower(*args, **kwds)


def nsdata_isspace(self, *args, **kwds):
    return bytes(self.bytes()).isspace(*args, **kwds)


def nsdata_istitle(self, *args, **kwds):
    return bytes(self.bytes()).istitle(*args, **kwds)


def isdata_isupper(self, *args, **kwds):
    return bytes(self.bytes()).isupper(*args, **kwds)


def nsdata_join(self, *args, **kwds):
    return bytes(self.bytes()).join(*args, **kwds)


def nsdata_ljust(self, *args, **kwds):
    return bytes(self.bytes()).ljust(*args, **kwds)


def nsdata_lower(self, *args, **kwds):
    return bytes(self.bytes()).lower(*args, **kwds)


def nsdata_lstrip(self, *args, **kwds):
    return bytes(self.bytes()).lstrip(*args, **kwds)


def nsdata_maketrans(cls, *args, **kwds):
    return bytes.maketrans(*args, **kwds)


def nsdata_partition(self, *args, **kwds):
    return bytes(self.bytes()).partition(*args, **kwds)


def nsdata_replace(self, *args, **kwds):
    return bytes(self.bytes()).replace(*args, **kwds)


def nsdata_rfind(self, *args, **kwds):
    return bytes(self.bytes()).rfind(*args, **kwds)


def nsdata_rindex(self, *args, **kwds):
    return bytes(self.bytes()).rindex(*args, **kwds)


def nsdata_rjust(self, *args, **kwds):
    return bytes(self.bytes()).rjust(*args, **kwds)


def nsdata_rpartition(self, *args, **kwds):
    return bytes(self.bytes()).rpartition(*args, **kwds)


def nsdata_rsplit(self, *args, **kwds):
    return bytes(self.bytes()).rsplit(*args, **kwds)


def nsdata_rstrip(self, *args, **kwds):
    return bytes(self.bytes()).rstrip(*args, **kwds)


def nsdata_split(self, *args, **kwds):
    return bytes(self.bytes()).split(*args, **kwds)


def nsdata_splitlines(self, *args, **kwds):
    return bytes(self.bytes()).splitlines(*args, **kwds)


def nsdata_startswith(self, *args, **kwds):
    return bytes(self.bytes()).startswith(*args, **kwds)


def nsdata_strip(self, *args, **kwds):
    return bytes(self.bytes()).strip(*args, **kwds)


def nsdata_swapcase(self, *args, **kwds):
    return bytes(self.bytes()).swapcase(*args, **kwds)


def nsdata_title(self, *args, **kwds):
    return bytes(self.bytes()).title(*args, **kwds)


def nsdata_translate(self, *args, **kwds):
    return bytes(self.bytes()).translate(*args, **kwds)


def nsdata_upper(self, *args, **kwds):
    return bytes(self.bytes()).upper(*args, **kwds)


def nsdata_zfill(self, *args, **kwds):
    return bytes(self.bytes()).zfill(*args, **kwds)


def nsdata__mul__(self, count):
    if count <= 0:
        return b""
    else:
        return bytes(self) * count


def nsdata__add__(self, other):
    return bytes(self) + other


def nsdata__radd__(self, other):
    return other + bytes(self)


addConvenienceForClass(
    "NSData",
    (
        ("__new__", staticmethod(nsdata__new__)),
        ("__len__", lambda self: self.length()),
        ("__str__", nsdata__str__),
        ("__getitem__", nsdata__getitem__),
        ("__bytes__", nsdata__bytes__),
        ("__mul__", nsdata__mul__),
        ("__rmul__", nsdata__mul__),
        ("__add__", nsdata__add__),
        ("__radd__", nsdata__radd__),
        ("capitalize", nsdata_capitalize),
        ("center", nsdata_center),
        ("count", nsdata_count),
        ("decode", nsdata_decode),
        ("endswith", nsdata_endswith),
        ("expandtabs", nsdata_expandtabs),
        ("find", nsdata_find),
        ("fromhex", classmethod(nsdata_fromhex)),
        ("hex", nsdata_hex),
        ("index", nsdata_index),
        ("isalnum", nsdata_isalnum),
        ("isalpha", nsdata_isalpha),
        ("isdigit", nsdata_isdigit),
        ("islower", nsdata_islower),
        ("isspace", nsdata_isspace),
        ("istitle", nsdata_istitle),
        ("isupper", isdata_isupper),
        ("join", nsdata_join),
        ("ljust", nsdata_ljust),
        ("lower", nsdata_lower),
        ("lstrip", nsdata_lstrip),
        ("maketrans", classmethod(nsdata_maketrans)),
        ("partition", nsdata_partition),
        ("replace", nsdata_replace),
        ("rfind", nsdata_rfind),
        ("rindex", nsdata_rindex),
        ("rjust", nsdata_rjust),
        ("rpartition", nsdata_rpartition),
        ("rsplit", nsdata_rsplit),
        ("rstrip", nsdata_rstrip),
        ("split", nsdata_split),
        ("splitlines", nsdata_splitlines),
        ("startswith", nsdata_startswith),
        ("strip", nsdata_strip),
        ("swapcase", nsdata_swapcase),
        ("title", nsdata_title),
        ("translate", nsdata_translate),
        ("upper", nsdata_upper),
        ("zfill", nsdata_zfill),
    ),
)


if sys.version_info[:2] >= (3, 7):

    def nsdata_isascii(self, *args, **kwds):
        return bytes(self.bytes()).isascii(*args, **kwds)

    addConvenienceForClass("NSData", (("isascii", nsdata_isascii),))


if sys.version_info[:2] >= (3, 9):

    def nsdata_removeprefix(self, *args, **kwds):
        return bytes(self.bytes()).removeprefix(*args, **kwds)

    def nsdata_removesuffix(self, *args, **kwds):
        return bytes(self.bytes()).removesuffix(*args, **kwds)

    addConvenienceForClass(
        "NSData",
        (
            ("removeprefix", nsdata_removeprefix),
            ("removesuffix", nsdata_removesuffix),
        ),
    )


def nsmutabledata__setitem__(self, item, value):
    self.mutableBytes()[item] = value


def nsmutabledata__delitem__(self, item):
    if isinstance(item, slice):
        start, stop, step = item.indices(len(self))
        if step == 1:
            self.replaceBytesInRange_withBytes_length_((start, stop - start), b"", 0)
        else:
            indices = range(start, stop, step)
            if step > 0:
                for idx in indices[::-1]:
                    self.replaceBytesInRange_withBytes_length_((idx, 1), b"", 0)
            else:
                for idx in indices:
                    self.replaceBytesInRange_withBytes_length_((idx, 1), b"", 0)
    else:
        item = operator.index(item)
        if item < 0:
            item += len(self)
            if item < 0:
                raise IndexError(f"{type(self).__name__} index out of range")
        self.replaceBytesInRange_withBytes_length_((item, 1), b"", 0)


def nsmutabledata__iadd__(self, iterable_of_ints):
    view = memoryview(iterable_of_ints)
    self.appendData_(view)

    return self


def nsmutabledata_extend(self, iterable_of_ints):
    try:
        view = memoryview(iterable_of_ints)
    except TypeError:
        self.appendData_(bytes(iterable_of_ints))
    else:
        self.appendData_(view)

    return self


def nsmutabledata__imul__(self, count):
    if count <= 0:
        self.clear()
        return self
    elif count == 1:
        return self

    else:
        # XXX: This is suboptimal...
        orig = self.copy()
        for _ in range(count - 1):
            self += orig
        return self


def nsmutabledata_reverse(self):
    value = bytes(self)[::-1]
    self.setData_(value)


def nsmutabledata_pop(self, index=-1):
    index = operator.index(index)

    if index < 0:
        index += len(self)
        if index < 0:
            raise IndexError(f"{type(self).__name__} index out of range")

    result = self[index]
    self.replaceBytesInRange_withBytes_length_((index, 1), b"", 0)
    return result


def nsmutabledata_remove(self, value):
    # XXX: Optimize me
    tmp = bytearray(self)
    tmp.remove(value)
    self.setData_(tmp)


def nsmutabledata_insert(self, index, value):
    # XXX: Optimize me
    tmp = bytearray(self)
    tmp.insert(index, value)
    self.setData_(tmp)


def nsmutabledata_append(self, value):
    value = operator.index(value)
    self += bytes([value])


def nsmutabledata_clear(self):
    self.setData_(b"")


addConvenienceForClass(
    "NSMutableData",
    (
        ("__setitem__", nsmutabledata__setitem__),
        ("__delitem__", nsmutabledata__delitem__),
        ("__iadd__", nsmutabledata__iadd__),
        ("__imul__", nsmutabledata__imul__),
        ("extend", nsmutabledata_extend),
        ("append", nsmutabledata_append),
        ("reverse", nsmutabledata_reverse),
        ("remove", nsmutabledata_remove),
        ("clear", nsmutabledata_clear),
        ("pop", nsmutabledata_pop),
        ("insert", nsmutabledata_insert),
    ),
)



__all__ = ()
import inspect

import objc
from objc._objc import _nameForSignature

basic_types = {
    objc._C_VOID: "void",
    objc._C_INT: "int",
    objc._C_UINT: "unsigned int",
    objc._C_LNG: "long",
    objc._C_ULNG: "unsigned long",
    objc._C_LNG_LNG: "long long",
    objc._C_ULNG_LNG: "unsigned long long",
    objc._C_FLT: "float",
    objc._C_DBL: "double",
    objc._C_SHT: "short",
    objc._C_USHT: "unsigned short",
    objc._C_CHR: "char",
    objc._C_UCHR: "unsigned char",
    objc._C_CHAR_AS_INT: "int8_t",
    objc._C_CHAR_AS_TEXT: "char",
    objc._C_UNICHAR: "UniChar",
    objc._C_BOOL: "bool",
    objc._C_NSBOOL: "BOOL",
    objc._C_ID: "id",
    objc._C_CLASS: "Class",
    objc._C_SEL: "SEL",
    objc._C_CHARPTR: "char*",
}

prefixes = {
    objc._C_IN: "in ",
    objc._C_OUT: "out ",
    objc._C_INOUT: "inout ",
    objc._C_CONST: "const ",
    objc._C_ONEWAY: "oneway ",
}


def describe_type(typestr):

    nm = basic_types.get(typestr)
    if nm is not None:
        return nm

    if typestr == b"^?":
        return "<FUNCTION>"
    elif typestr == b"@?":
        return "<BLOCK>"

    if typestr.startswith(objc._C_PTR):
        nm = _nameForSignature(typestr)
        if nm is not None:
            return nm
        return describe_type(typestr[1:]) + "*"

    if typestr[:1] in prefixes:
        return prefixes[typestr[:1]] + describe_type(typestr[1:])

    if typestr.startswith(objc._C_STRUCT_B):
        nm = _nameForSignature(typestr)
        if nm is not None:
            return nm

        typestr = typestr[1:]
        idx = typestr.find(b"=")
        if idx == -1:
            return "struct <?>"

        else:
            nm = typestr[:idx]
            if not nm:
                return "struct <?>"
            nm = nm.decode("utf-8").lstrip("_")
            if "." in nm:
                return nm.rsplit(".", 1)[-1]
            return nm

    if typestr.startswith(objc._C_VECTOR_B):
        base_type = basic_types[typestr[-2:-1]]
        if base_type.startswith("unsigned "):
            base_type = "u" + base_type.split()[-1]

        count = typestr[1:-2]
        return f"simd_{base_type}{count.decode()}"

    if typestr.startswith(objc._C_ARY_B):
        typestr = typestr[1:]
        d = b""
        while typestr[:1].isdigit():
            d += typestr[:1]
            typestr = typestr[1:]

        return "{}[{}]".format(describe_type(typestr), d.decode("utf-8"))

    if typestr.startswith(objc._C_UNION_B):
        typestr = typestr[1:]
        idx = typestr.find(b"=")
        if idx == -1:
            return "union <?>"

        else:
            nm = typestr[:idx]
            if not nm:
                nm = b"<?>"
            return "union {}".format(nm.decode("utf-8"))

    return "<?>"


def describe_callable(callable_object):
    if not hasattr(callable_object, "__name__") or not hasattr(
        callable_object, "__metadata__"
    ):
        return None

    name = callable_object.__name__
    try:
        metadata = callable_object.__metadata__()
    except objc.internal_error:
        return None

    return describe_callable_metadata(
        name, metadata, ismethod=isinstance(callable_object, objc.selector)
    )


def describe_callable_metadata(name, metadata, offset="", ismethod=False):
    arg_info = []
    if ismethod:
        arg_offset = 2

        name_parts = name.split(":")
        hdr_name = []
        if len(metadata["arguments"]) > arg_offset:
            for idx, (nm, info) in enumerate(
                zip(name_parts, metadata["arguments"][arg_offset:])
            ):
                if hdr_name:
                    hdr_name.append(" ")
                hdr_name.append(nm)
                hdr_name.append(":(")
                hdr_name.append(describe_type(info["type"]))
                hdr_name.append(")arg%d" % (idx,))

                if info["type"][:1] in prefixes and info["type"][:1] not in (
                    objc._C_ONEWAY,
                    objc._C_CONST,
                ):
                    arg_info.append((idx, info))
                elif info.get("printf_format"):
                    arg_info.append((idx, info))
                elif info.get("callable"):
                    arg_info.append((idx, info))

        else:
            hdr_name.append(name)

        if metadata.get("variadic"):
            hdr_name.append(", ...")

        header = "{} ({}){};".format(
            "+" if metadata["classmethod"] else "-",
            describe_type(metadata["retval"]["type"]),
            "".join(hdr_name),
        )
    else:
        hdr_name = []
        arg_offset = 0
        for idx, info in enumerate(metadata["arguments"]):
            if idx != 0:
                hdr_name.append(", ")
            hdr_name.append(describe_type(info["type"]))
            hdr_name.append(" arg%d" % (idx,))
            if info["type"][:1] in prefixes and info["type"][:1] not in (
                objc._C_ONEWAY,
                objc._C_CONST,
            ):
                arg_info.append((idx, info))
            elif info.get("printf_format"):
                arg_info.append((idx, info))
            elif info.get("callable"):
                arg_info.append((idx, info))
        if metadata.get("variadic"):
            hdr_name.append(", ...")

        if not hdr_name:
            hdr_name.append("void")

        header = "{} {}({});".format(
            describe_type(metadata["retval"]["type"]),
            name,
            "".join(hdr_name),
        )

    result = [header]
    if metadata.get("suggestion"):
        result.append("")
        result.append("WARNING: {}".format(metadata["suggestion"]))

    if arg_info:
        result.append("")
        for idx, info in arg_info:
            if info.get("printf_format"):
                result.append("arg%d: %%-style format string" % (idx,))
                continue

            elif info.get("callable"):
                result.append(
                    "arg%d: %s"
                    % (
                        idx,
                        describe_callable_metadata(
                            "callback", info["callable"], offset="    " + offset
                        ),
                    )
                )
                continue

            else:
                arg = info.get("c_array_length_in_arg")
                if arg is not None:
                    if isinstance(arg, tuple):
                        result.append(
                            "arg%d: array with length on input in arg%d, "
                            "and output in arg%d"
                            % (idx, arg[0] - arg_offset, arg[1] - arg_offset)
                        )
                    else:
                        if info.get("c_array_length_in_result"):
                            result.append(
                                "arg%d: array with length on input in arg%d, "
                                "and output in return value" % (idx, arg - arg_offset)
                            )
                        else:
                            result.append(
                                "arg%d: array with length in arg%d"
                                % (idx, arg - arg_offset)
                            )
                    continue

                if info.get("c_array_length_in_result"):
                    result.append("arg%d: array with length in return value" % (idx,))
                    continue

                if info.get("c_array_of_fixed_length"):
                    result.append(
                        "arg%d: array with length %d"
                        % (idx, info.get("c_array_of_fixed_length"))
                    )
                    continue

                if info.get("c_array_of_variable_length"):
                    result.append("arg%d: array with unknown length" % (idx,))
                    continue

                if info.get("c_array_delimited_by_null"):
                    result.append(
                        "arg%d: array (will be NULL terminated in C)" % (idx,)
                    )
                    continue

                result.append(
                    "arg%d: pass-by-reference %sargument"
                    % (idx, prefixes.get(info["type"][:1]))
                )

    if len(metadata["arguments"]) > arg_offset:
        if metadata.get("variadic") and metadata.get("c_array_delimited_by_null"):
            if not arg_info:
                result.append("")

            result.append(
                "Variadic arguments form an array of C type %s"
                % (describe_type(metadata["arguments"][-1]["type"]),)
            )

    if metadata.get("__doc__", None):
        result.append("\n")
        result.append(str(metadata["__doc__"]))

    return ("\n" + offset).join(result).replace("\n" + offset + "\n", "\n\n")


objc.options._callable_doc = describe_callable


def callable_signature(callable_object):
    # Create an inspect.Signature for an PyObjC callable
    # both objc.function and objc.native_selector only support positional
    # arguments, and not keyword arguments.
    if not hasattr(callable_object, "__name__") or not hasattr(
        callable_object, "__metadata__"
    ):
        return None

    try:
        metadata = callable_object.__metadata__()
    except objc.internal_error:
        # This can happen with some private methods with undocumented
        # characters in type encodings
        return None

    ismethod = isinstance(callable_object, objc.selector)
    isimp = isinstance(callable_object, objc.IMP)

    if ismethod:
        # Skip 'self' and 'selector' implicit arguments
        args = metadata["arguments"][2:]
    elif isimp:
        # Ignore the 'selector' argument
        args = (metadata["arguments"][0],) + metadata["arguments"][2:]
    else:
        args = metadata["arguments"]

    parameters = []
    for idx, _arg in enumerate(args):
        p_name = "arg%d" % (idx,)
        parameters.append(inspect.Parameter(p_name, inspect.Parameter.POSITIONAL_ONLY))

    return inspect.Signature(parameters)


objc.options._callable_signature = callable_signature



import objc
import functools
import sys


def _structConvenience(structname, structencoding):
    def makevar(cls, name=None):
        if name is None:
            return objc.ivar(type=structencoding)
        else:
            return objc.ivar(name=name, type=structencoding)

    makevar.__name__ = structname
    makevar.__doc__ = f"Create *ivar* for type encoding {structencoding!r}"
    if hasattr(objc.ivar, "__qualname__"):  # pragma: no branch
        makevar.__qualname__ = objc.ivar.__qualname__ + "." + structname

    objc.ivar._add_attribute(sys.intern(structname), classmethod(makevar))


# Fake it for basic C types
_structConvenience("bool", objc._C_BOOL)
_structConvenience("char", objc._C_CHR)
_structConvenience("int", objc._C_INT)
_structConvenience("short", objc._C_SHT)
_structConvenience("long", objc._C_LNG)
_structConvenience("long_long", objc._C_LNG_LNG)
_structConvenience("unsigned_char", objc._C_UCHR)
_structConvenience("unsigned_int", objc._C_UINT)
_structConvenience("unsigned_short", objc._C_USHT)
_structConvenience("unsigned_long", objc._C_ULNG)
_structConvenience("unsigned_long_long", objc._C_ULNG_LNG)
_structConvenience("float", objc._C_FLT)
_structConvenience("double", objc._C_DBL)
_structConvenience("BOOL", objc._C_NSBOOL)
_structConvenience("UniChar", objc._C_UNICHAR)
_structConvenience("char_text", objc._C_CHAR_AS_TEXT)
_structConvenience("char_int", objc._C_CHAR_AS_INT)

_orig_createStructType = objc.createStructType


@functools.wraps(objc.createStructType)
def createStructType(name, typestr, fieldnames, doc=None, pack=-1):
    result = _orig_createStructType(name, typestr, fieldnames, doc, pack)
    _structConvenience(name, result.__typestr__)
    return result


objc.createStructType = createStructType


_orig_registerStructAlias = objc.registerStructAlias


@functools.wraps(objc.registerStructAlias)
def registerStructAlias(typestr, structType):
    return _orig_registerStructAlias(typestr, structType)


def createStructAlias(name, typestr, structType):
    result = _orig_registerStructAlias(typestr, structType)
    _structConvenience(name, result.__typestr__)
    return result


objc.createStructAlias = createStructAlias
objc.registerStructAlias = registerStructAlias



"""Signals.py -- dump a python stacktrace if something bad happens.

          DO NOT USE THIS MODULE IN PRODUCTION CODE

This module has two functions in its public API:

- dumpStackOnFatalSignal()
  This function will install signal handlers that print a stacktrace and
  then reraise the signal.

- resetFatalSignals()
  Restores the signal handlers to the state they had before the call to
  dumpStackOnFatalSignal.

This module is not designed to provide fine grained control over signal
handling. Nor is it intended to be terribly robust. It may give useful
information when your program gets unexpected signals, but it might just
as easily cause a crash when such a signal gets in.

          DO NOT USE THIS MODULE IN PRODUCTION CODE
"""

import os
import signal
import traceback
import warnings
import sys

__all__ = ["dumpStackOnFatalSignal", "resetFatalSignals"]

warnings.warn(
    "PyObjCTools.Signals is deprecated and will be removed in PyObjC 9",
    DeprecationWarning,
)


originalHandlers = None


def dumpHandler(signum, frame):
    """
    the signal handler used in this module: print a stacktrace and
    then re-raise the signal
    """
    resetFatalSignals()
    print("*** Handling fatal signal '%d'." % signum, file=sys.stderr)
    traceback.print_stack(frame)
    print("*** Restored handlers and resignaling.", file=sys.stderr)
    os.kill(os.getpid(), signum)


def installHandler(sig):
    """
    Install our signal handler for a signal. The original handler
    is saved in 'originalHandlers'.
    """
    originalHandlers[sig] = signal.signal(sig, dumpHandler)


def dumpStackOnFatalSignal():
    """
    Install signal handlers that might print a useful stack trace when
    this process receives a fatal signal.

    NOTE: See module docstring
    """

    global originalHandlers
    if not originalHandlers:
        originalHandlers = {}
        installHandler(signal.SIGQUIT)
        installHandler(signal.SIGILL)
        installHandler(signal.SIGTRAP)
        installHandler(signal.SIGABRT)
        installHandler(signal.SIGEMT)
        installHandler(signal.SIGFPE)
        installHandler(signal.SIGBUS)
        installHandler(signal.SIGSEGV)
        installHandler(signal.SIGSYS)


def resetFatalSignals():
    """
    Restore the original signal handlers
    """
    global originalHandlers
    if originalHandlers:
        for sig in originalHandlers:
            signal.signal(sig, originalHandlers[sig])
        originalHandlers = None
